/*
 * Copyright (c) Oona Räisänen
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 */
#ifndef GROUPS_H_
#define GROUPS_H_

#include <chrono>
#include <iostream>
#include <map>
#include <memory>
#include <set>
#include <string>
#include <vector>

#include <json/json.h>

#include "config.h"
#include "src/common.h"
#include "src/options.h"
#include "src/rdsstring.h"
#include "src/tmc/tmc.h"
#include "src/util.h"

namespace redsea {

enum eBlockNumber {
  BLOCK1, BLOCK2, BLOCK3, BLOCK4
};

enum class GroupTypeVersion {
  A, B
};

enum class Offset {
  A, B, C, Cprime, D, invalid
};

class GroupType {
 public:
  explicit GroupType(uint16_t type_code = 0x00);

  bool operator==(const GroupType& other);

  std::string str() const;

  uint16_t number;
  GroupTypeVersion version;
};

bool operator<(const GroupType& type1, const GroupType& type2);

class Group {
 public:
  Group();
  uint16_t block(eBlockNumber block_num) const;
  bool has(eBlockNumber block_num) const;
  bool empty() const;
  GroupType type() const;
  bool has_type() const;
  uint16_t pi() const;
  uint8_t bler() const;
  uint8_t num_errors() const;
  bool has_pi() const;
  bool has_bler() const;
  bool has_time() const;
  std::chrono::time_point<std::chrono::system_clock> rx_time() const;
  void set(eBlockNumber block_num, uint16_t data, bool had_errors = false);
  void set_c_prime(uint16_t data, bool had_errors = false);
  void disable_offsets();
  void set_time(std::chrono::time_point<std::chrono::system_clock> t);
  void set_bler(uint8_t bler);
  void set_channel(int which_channel);

 private:
  GroupType type_;
  uint16_t pi_;
  std::vector<bool> has_block_;
  std::vector<uint16_t> block_;
  std::vector<bool> block_had_errors_;
  std::chrono::time_point<std::chrono::system_clock> time_received_;
  uint8_t bler_;
  bool has_type_;
  bool has_pi_;
  bool has_c_prime_;
  bool has_bler_;
  bool has_time_;
  bool no_offsets_;
};

class Station {
 public:
  Station();
  Station(uint16_t _pi, const Options& options, int which_channel);
  void UpdateAndPrint(const Group& group, std::ostream* stream);
  uint16_t pi() const;

 private:
  void DecodeBasics(const Group& group);
  void DecodeType0(const Group& group);
  void DecodeType1(const Group& group);
  void DecodeType2(const Group& group);
  void DecodeType3A(const Group& group);
  void DecodeType4A(const Group& group);
  void DecodeType6(const Group& group);
  void DecodeType14(const Group& group);
  void DecodeType15B(const Group& group);
  void DecodeODAGroup(const Group& group);
  void AddAltFreq(uint8_t);
  void UpdatePS(int pos, int char1, int char2);
  void UpdateRadioText(int pos, int char1, int char2);
  void ParseRadioTextPlus(const Group& group);
  uint16_t pi_;
  Options options_;
  int which_channel_;
  RDSString ps_;
  RDSString radiotext_;
  int radiotext_ab_;
  int pty_;
  bool is_tp_;
  bool is_ta_;
  bool is_music_;
  int pin_;
  int ecc_;
  int cc_;
  int tmc_id_;
  int ews_channel_;
  int lang_;
  bool linkage_la_;
  std::string clock_time_;
  bool has_country_;
  std::map<GroupType, uint16_t> oda_app_for_group_;
  bool has_radiotext_plus_;
  bool radiotext_plus_cb_;
  uint16_t radiotext_plus_scb_;
  uint16_t radiotext_plus_template_num_;
  bool radiotext_plus_toggle_;
  bool radiotext_plus_item_running_;
  std::map<uint16_t, RDSString> eon_ps_names_;
  std::map<uint16_t, AltFreqList> eon_alt_freqs_;
  bool last_block_had_pi_;
  AltFreqList alt_freq_list_;

  int pager_pac_;
  int pager_opc_;
  int pager_tng_;
  int pager_ecc_;
  int pager_ccf_;
  int pager_interval_;

  Json::StreamWriterBuilder writer_builder_;
  std::unique_ptr<Json::StreamWriter> writer_;
  Json::Value json_;

#ifdef ENABLE_TMC
  tmc::TMC tmc_;
#endif
};

struct RTPlusTag {
  uint16_t content_type;
  uint16_t start;
  uint16_t length;
};

void PrintHexGroup(const Group& group, std::ostream* stream,
                   const std::string& time_format);

}  // namespace redsea
#endif  // GROUPS_H_
